function [p, t,df] = mattest(X1, X2, varargin)
%MATTEST Two-tailed t-test for identifying differentially expressed genes
% from two different conditions/phenotypes.
%
%   P = MATTEST(X1,X2) performs a t-test for differential expression with a
%   standard two-sided t-test on every row (gene) of the expression
%   matrices, and returns the results of the p-values for every gene. The
%   data are assumed to be normally distributed in each class with equal
%   variances. X1 and X2 must have the same number of rows. 
%
%   [P, T] = MATTEST(...) returns the t-scores T for each gene.
%
%   [P, T, DF] = MATTEST(...) returns the degrees of freedom DF of the
%   test.
%
%   MATTEST(...,'PERMUTE', P) perform permutation tests to if P is set to
%   TRUE.  The default number of permutation is 1000. P can also be the
%   number of permutations if is set to be greater than 2. The default is
%   FALSE.
% 
%   MATTEST(...,'SHOWHIST',TRUE) displays histograms of the p-value
%   distribution and t-score distribution. 
%
%   MATTEST(...,'SHOWPLOT',TRUE) displays normal t-scores quantile plot.
%   The points with t-score greater than the (1 - 1/(2N)) quantile or less
%   than the 1/(2N) quantile are marked with red circles. N is total number
%   of genes in the t-test.
% 
%   MATTEST(...,'LABELS',LABELS) allows you to specify a cell array of gene
%   labels or names for each row of the input data. If LABELS are defined,
%   then clicking on a data point on the plot will show the LABEL
%   corresponding to that data point.
% 
%   Example:
%       load prostatecancerexpdata; 
%       [p,t]=mattest(dependentData,independentData, 'showplot', true);
% 
%       % Permutation t-tests.
%       [p,t]=mattest(dependentData,independentData, 'permute', true,...
%                           'Showhist', true', 'showplot', true);
%   
%   See also GCRMA, MAFDR, MAIRPLOT, MALOGLOG, MAPCAPLOT, MAVOLCANOPLOT,
%   RMASUMMARY.

% Copyright 2005-2006 The MathWorks, Inc.
% $Revision: 1.1.10.4 $   $Date: 2007/01/12 01:27:32 $ 

% References: 
%  [1] W. Huber, A. von Heydebreck, H. Sltmann, A. Poustka, M. Vingron;
%      "Variance stabilization applied to microarray data calibration and
%      to the quantification of differential expression", Bioinformatics
%      (2002) 18 Suppl1 S96-S104.

%set defaults
qplotFlag = false;
histFlag = false;
labels = [];
nperms = 0;

if nargin < 2
    error('Bioinfo:TooFewInputArguments',...
        'Too few input arguments to %s.',mfilename);
end

if ~isnumeric(X1) || ~isreal(X1) || ~isnumeric(X2) || ~isreal(X2)
   error('Bioinfo:mattest:ExpressionValuesNotNumericAndReal',...
         'Expression values must be numeric and real.') 
end

% validate required inputs 
if size(X1,1)~=size(X2,1)
    error('Bioinfo:mattest:differentSize',...
        'X1 and X2 must have same number of rows.')
end

if  nargin > 2
    if rem(nargin,2) == 1
        error('Bioinfo:mattest:IncorrectNumberOfArguments',...
            'Incorrect number of arguments to %s.',mfilename);
    end
    okargs = {'permute', 'showhist', 'showplot', 'labels'};
    for j=1:2:nargin-2
        pname = varargin{j};
        pval = varargin{j+1};
        k = strmatch(lower(pname), okargs);%#ok
        if isempty(k)
            error('Bioinfo:mattest:UnknownParameterName',...
                'Unknown parameter name: %s.',pname);
        elseif length(k)>1
            error('Bioinfo:mattest:AmbiguousParameterName',...
                'Ambiguous parameter name: %s.',pname);
        else
            switch(k)
                case 1 % permutation
                    if opttf(pval)
                        if isnumeric(pval)
                            if isscalar(pval)
                                nperms = ceil(double(pval));
                                if nperms <= 2
                                    error('Bioinfo:mattest:InvalidPermutationNumber',...
                                        'Number of permutation tests must be greater than 2.');
                                end
                            else
                                nperms = 1000;
                                warning('Bioinfo:mattest:PermuteNumberNoScalar',...
                                    'PERMUTE must be a single number or a logical. \nRun permutation test 1000 times by default.')
                            end
                        else
                            nperms = 1000;
                        end
                    else
                        nperms = 0;
                    end
                case 2 % showhist
                    histFlag = opttf(pval);
                    if isempty(histFlag)
                        error('Bioinfo:InputOptionNotLogical',...
                            '%s must be a logical value, true or false.',...
                            upper(char(okargs(k))));
                    end
                 case 3 % show quantile plot
                    qplotFlag = opttf(pval);
                    if isempty(qplotFlag)
                        error('Bioinfo:InputOptionNotLogical',...
                            '%s must be a logical value, true or false.',...
                            upper(char(okargs(k))));
                    end
                case 4 % labels
                    labels = pval;
            end
        end
    end
end

% Check the labels size
if(~isempty(labels))
    if ~isvector(labels) || numel(labels) < size(X1,1)
        warning('Bioinfo:mattest:MismatchedLabels',...
            ['The number of labels did not match the number of rows in input data.',...
            '\nThe quantile plot of t was not able to show the datatip.']);
        labels = [];
    end
    
    labels = labels(:);
end

% do ttest
if nperms ~= 0
    [p, t, df] = permutationtests(X1, X2, nperms);
else
    [h_dum,p,ci_dum,stats] = ttest2(X1, X2, [],[],[],2); %#ok
    t = stats.tstat;
    df = stats.df;
end

if qplotFlag  
    hqpfig = plotquantilet(t, df, labels); %#ok
end

if histFlag
    hdfig = plothistogram(t, p); %#ok
end

%*********** Callbacks and helper functions ***************
function [p, t, df] = permutationtests(X1, X2, nperms)
% Number of samples
ndata1 = size(X1, 2);
ndata2 = size(X2, 2);
nsamples = ndata1 + ndata2;
samples = [X1 X2];
m = size(X1, 1);

% Compute observed t statitics
grpmean = [mean(X1,2) mean(X2,2)];
grpvarm = [var(X1,[],2)./ndata1 var(X2,[],2)./ndata2];
t = diff(grpmean,[],2) ./ sqrt(sum(grpvarm,2));

% Conduct a permutation test to create a sample from the null distribution
% of the t statistics. Permute the transcripts at random, compute t-scores
% among the random partitions of the actual samples, repeat for a large
% number permutations.
% Bin the samples
edge = max(findedge(t), 20);
t0edges = linspace(-edge, edge, getUpperGeneNumber(m)+1)'; % #bins odd -> edge at zero
t0bins = t0edges(1:end-1) + diff(t0edges)/2;
t0 =zeros(size(t0edges));

for i = 1:nperms
    [group1, group2] = crossvalind('LeaveMOut', nsamples, ndata2);
    grpmean = [mean(samples(:, group1), 2) mean(samples(:, group2), 2)];
    grpvar = [var(samples(:, group1), [], 2)./ndata1 var(samples(:, group2), [], 2)./ndata2];
    t0rep = diff(grpmean, [], 2) ./sqrt(sum(grpvar, 2));
    t0 = t0 + histc(t0rep, t0edges);
end

t0 = t0(1:end-1); % drop the degenetated uppermost bin

% Estimate the null distribution of |t| and compute its empirical
% cumulative distribution function.
[f0, x0] = ecdf(abs(t0bins), 'freq', t0);
x0(1) = 0;

% Estimate p-values corresponding to truly null hypotheses for the observed
% t-scores from the t-test
pp=interp1(x0, f0, 'pchip', 'pp');
p=abs(1-ppval(pp, abs(t)));

% % p = 1 - interp1q(x0, f0, abs(t));

df = ndata1 + ndata2 -2;
if ~isscalar(t)
    df = repmat(df,size(t));
end

%----------------------------------------------------------
function hqpfig = plotquantilet(tscore, df, labels)
% do quantile plot
% df - degree of freedom.

hqpfig = figure('Units', 'Normalized',...
    'Tag', 'ttestquantileplot',...
    'Name', 'Normal Quantile Plot of t',...
    'numbertitle','off', ...
    'Visible', 'off');

appdata = localGetAppData(hqpfig);

% Remove the NaNs 
goodidx = ~isnan(tscore);
tscore = tscore(goodidx);
df = df(goodidx);

if ~isempty(labels)
    labels = labels(goodidx);
end

[appdata.tscore, ord] = sort(tscore);
if ~isempty(labels)
    appdata.labels = labels(ord);
    
    if ~ischar(labels) && isnumeric(labels)
        appdata.labels = num2str(appdata.labels(:));
        appdata.labels = cellstr(appdata.labels);
    end
    
end
% number of genes
N = numel(tscore);

% get quatile function of t stats
Xq = (1:N)';
Xq = (Xq - 0.5)/N;

appdata.tquantile = tinv(Xq, df);

appdata.xlim = [min(appdata.tquantile), max(appdata.tquantile)];

appdata.haxes = axes;
set(appdata.haxes, 'Parent', hqpfig,...
    'Box', 'on',...
    'GridLineStyle', '--');

if appdata.tquantile ~= 0
    set(appdata.haxes, 'XLimMode', 'manual',...
    'YLimMode', 'auto',...
    'XLim',appdata.xlim);
end

%up-regulated threshold
upThd = tinv( 1- 0.5/N, df);
upIdx = find(appdata.tscore > upThd);
%up-regulated threshold
dnThd = tinv( 0.5/N, df);
dnIdx = find(appdata.tscore < dnThd);

appdata.hplot = plot(appdata.tquantile, appdata.tscore, 'b.',...
    appdata.tquantile(upIdx), appdata.tscore(upIdx),'ro',...
    appdata.tquantile(dnIdx), appdata.tscore(dnIdx), 'ro',...
    appdata.tquantile, appdata.tquantile, 'k-.');

set(appdata.hplot, 'parent', appdata.haxes);
xlabel('Theoretical quantile')
ylabel('Sample quantile')

% If labels exist, set up a buttondown function
if ~isempty(appdata.labels)
    set(appdata.haxes,'ButtonDownFcn',@clickonplot);
    set(appdata.hplot,'ButtonDownFcn',@clickonplot);
end

localSetAppData(hqpfig,appdata);

set(hqpfig, 'Visible', 'on');

%**************************************
function hdfig = plothistogram(tscore, pvalue)

   hdfig = figure('Units', 'Normalized',...
        'Tag', 'ttesthistogram',...
        'Name', 'Histograms of t-test Results',...
        'numbertitle','off', ...
        'Visible', 'off');
    subplot(1,2,1);
    bw = .5;
    edge = findedge(tscore);
    edges = -edge:bw:edge;
%     hgt1 = histc(tscore,edges) ./ (length(tscore)*bw);
    hgt1 = histc(tscore,edges);
    bar(edges,hgt1,'histc');
    h = findobj(gca,'Type','patch');
    set(h,'FaceColor',[1.0 0.3 0.3], 'EdgeColor', 'w')
    xlim([-11 11]);
    title('t-scores')
    xlabel('t-score');
    ylabel('Frequency')
    
    subplot(1,2,2)
    hist(pvalue, 0.025:0.05:.975);
    h = findobj(gca,'Type','patch');
    set(h,'FaceColor',[0.5 0.5 0.8], 'EdgeColor', 'w') %[0.3 0 1]
    title('p-values')
    xlabel('p-value');
    ylabel('Frequency')
    
    set(hdfig, 'visible', 'on')

%**************************************
function clickonplot(varargin) %#ok
% callback function highlights selected element and displays label.
[hObj,hfig] = gcbo; %#ok

appdata = localGetAppData(hfig);
set(hfig,'Doublebuffer','on');

xvals = appdata.tquantile;
yvals = appdata.tscore;

point = get(appdata.haxes,'CurrentPoint');
px = point(1,1);
py = point(1,2);

% find the closest point
[v, index] = min(((xvals - px).^2)./(px^2) + ((yvals - py).^2)./(py ^2)); %#ok

%highlight the point and set the title
if ~isempty(index)
    hHighlight = line(xvals(index),yvals(index),...
            'color',[0.3, 0.9 0.0],'marker','d','Tag','THighlight', 'linewidth', 2);%#ok

    % create a new text object -- start with it invisible
    htext = text(px, py , appdata.labels(index), 'visible','off','interpreter','none');

    set(htext,	'fontsize', 8, 'Tag','TDataTip');
    setlabelposition(htext, appdata.haxes)
    % show the text
    set(htext, 'Visible','on')
    set(hfig,  'WindowButtonUpFcn', @cleanUpLabel);
    localSetAppData(hfig,appdata);
end

%********************************************************
function cleanUpLabel(varargin) %#ok
% cleanUpLabel callback function to remove label from plot
% get the handles to the figure, plot and axis

hFig = gcbf;

% delete the old label if it exists
oldLabel = findobj(hFig,'Tag','TDataTip');
if ~isempty(oldLabel)
    delete(oldLabel);
end

% delete the old label if it exists
oldHighlight = findobj(hFig,'Tag','THighlight');
if ~isempty(oldHighlight)
    delete(oldHighlight);
end

%***************************************************************
function [appdata] = localGetAppData(hfig)

if isappdata(hfig,'MAQuantileTPlot')
    appdata = getappdata(hfig,'MAQuantileTPlot');
else
    appdata = guihandles(hfig);
    
    appdata.tscore = [];
    appdata.labels = [];
    
    appdata.tquantile = [];

    appdata.xlim = [];
    
    % Do the plot
    appdata.haxes = [];
    appdata.hplot = [];
end
%----------------------------------------------------------------
function localSetAppData(hfig,appdata)
setappdata(hfig,'MAQuantileTPlot',appdata);

function setlabelposition(htext, haxes)
% If the text extent outside the axis, move it inside

% Get text extent,a nd axes range
ext = get(htext, 'Extent');
xrange = get(haxes,'Xlim');
yrange = get(haxes,'Ylim');

if ext(1)+ext(3) > xrange(2)
    set(htext, 'HorizontalAlignment','right')
elseif ext(2) > yrange(2)
    set(htext, 'VerticalAlignment','top')
else
    return;
end

%-----------------------------------------------
function n = getUpperGeneNumber(n)
% Return the rounded upper gene number
% for example, 3893 genes, return 4000

fn = floor(log10(n));
cn = ceil(log10(n));
if cn-fn == 0
    return;
end

m = ceil(n/(10^fn));

n = m*10^fn;

%---------------------------------------------------
function edge = findedge(t)
% return the rounded max |t| 

maxt = getUpperGeneNumber(abs(max(t)));
mint = getUpperGeneNumber(abs(min(t)));
edge = max(maxt, mint);

