/* reliability.h: functions creating an interface to functions from
 * the GNU multiprecission library, thus making it easy to replace
 * e.g. rationals with floats, or even with non-GMP types like
 * doubles.
 */
/* Rune Lyngs, February 2003 */

#include <stdio.h>
#include <stdlib.h>
#include "gmp_interface.h"
#ifdef USE_GMP
#include "gmp.h"
#endif

/* init(X): prepares X for use */
void gmpi_init(mp_type X)
{
#ifdef USE_GMP
  mpq_init(X);
#endif
}

/* clear(X): decomissions use of X */
void gmpi_clear(mp_type X)
{
#ifdef USE_GMP
  mpq_clear(X);
#endif
}

/* set(ROP, OP): assigns value of OP to ROP */
void gmpi_set(mp_type ROP, mp_type OP)
{
#ifdef USE_GMP
  mpq_set(ROP, OP);
#else
  ROP[0] = OP[0];
#endif
}

/* set_ui(ROP, OP): assigns value of OP to ROP */
void gmpi_set_ui(mp_type ROP, unsigned long int OP)
{
#ifdef USE_GMP
  mpq_set_ui(ROP, OP, 1);
#else
  ROP[0] = (double)OP;
#endif
}

void gmpi_set_si(mp_type ROP, signed long int OP1, unsigned long int OP2)
{
#ifdef USE_GMP
  mpq_set_si(ROP, OP1, OP2);
  mpq_canonicalize(ROP);
#else
  ROP[0] = (double)OP1 / (double)OP2;
#endif
}

/* set_d(ROP, OP): assigns value of OP to ROP */
void gmpi_set_d(mp_type ROP, double OP)
{
#ifdef USE_GMP
  mpq_set_d(ROP, OP);
#else
  ROP[0] = OP;
#endif
}

/* init_set_ui(ROP, OP): simultaneously prepares OP for use and set it to OP */
void gmpi_init_set_ui(mp_type ROP, unsigned long int OP)
{
#ifdef USE_GMP
  mpq_init(ROP);
  mpq_set_ui(ROP, OP, 1);
#else
  ROP[0] = (double)OP;
#endif
}

/* add(ROP, OP1, OP2): assigns sum of OP1 and OP2 to ROP */
void gmpi_add(mp_type SUM, mp_type OP1, mp_type OP2)
{
#ifdef USE_GMP
  mpq_add(SUM, OP1, OP2);
#else
  SUM[0] = OP1[0] + OP2[0];
#endif
}

/* inc(OP): add 1 to value of OP */
void gmpi_inc(mp_type OP)
{
#ifdef USE_GMP
  mpz_add(mpq_numref(OP), mpq_numref(OP), mpq_denref(OP));
#else
  OP[0] += 1;
#endif
}

/* sub(ROP, OP1, OP2): assigns difference between OP1 and OP2 to ROP */
void gmpi_sub(mp_type SUM, mp_type OP1, mp_type OP2)
{
#ifdef USE_GMP
  mpq_sub(SUM, OP1, OP2);
#else
  SUM[0] = OP1[0] - OP2[0];
#endif
}

/* inv(OP): invert OP, i.e. set OP = 1 / OP */
void gmpi_inv(mp_type OP)
{
#ifdef USE_GMP
  mpq_inv(OP, OP);
#else
  OP[0] = 1.0 / OP[0];
#endif
}

/* mul(ROP, OP1, OP2): assigns product of OP1 and OP2 to ROP */
void gmpi_mul(mp_type ROP, mp_type OP1, mp_type OP2)
{
#ifdef USE_GMP
  mpq_mul(ROP, OP1, OP2);
#else
  ROP[0] = OP1[0] * OP2[0];
#endif
}

/* div(ROP, OP1, OP2): assigns quotient of OP1 and OP2 to ROP */
void gmpi_div(mp_type ROP, mp_type OP1, mp_type OP2)
{
#ifdef USE_GMP
  mpq_div(ROP, OP1, OP2);
#else
  ROP[0] = OP1[0] / OP2[0];
#endif
}

/* get_d(OP): converts OP to double */
double gmpi_get_d(mp_type OP)
{
#ifdef USE_GMP
  return mpq_get_d(OP);
#else
  return OP[0];
#endif
}

void gmpi_print(mp_type OP)
{
#ifdef USE_GMP
  mpq_out_str(stdout, 10, OP);
#else
  printf("%f", OP[0]);
#endif
}
