/* iterated_global_normalisation.c: Implementation of functions for
 * computing normalising functions based on an iterative
 * scheme. Starting with all points having equal weight, in subsequent
 * iterations the weight of a particular point is adjusted according
 * to its distance to the normalising function obtained in the
 * previous iteration, such that the weight trails off exponentially
 * in the squared distance.
 */
/* Rune Lyngs, March 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "iterated_global_normalisation.h"
#include "rankselect.h"
#include "wrappers.h"

static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "iterated_global_normalisation");
#else
  return wmalloc(n);
#endif
}

/* cmp_doublep(x, y): compares the doubles pointed to by x and y */
static int cmp_doublep(double *x, double *y)
{
  return *x < *y;
}

/* weights(f, eval_function, m, x, y, weight): Compute weights for the
 * m points given by x and y based on the distance in s standard
 * deviations y is from the normalisation function f at x, where
 * eval_function is used to evaluate f at x; if no function for
 * computing weights is specified, i.e. weight is NULL, the default
 * formula for computing the weight is 2 raised to minus this distance
 * squared. Returns the weights as an array of doubles.
 */
static double *weights(void *f, _IGN_evalfunction eval_function, int m,
		       double *x, double *y, _IGN_weight weight)
{
  double *w = (double *)xmalloc(m * sizeof(double));
  int i;
  double mean = 0, sd = 0, median = 0;

  if (weight == NULL){
    /* Use default method for assigning weights */
    /* Compute value of f at each x, and sum of square differences to
     * corresponding y values.
     */
    for (i = 0; i < m; i++){
      w[i] = fabs(eval_function(f, x[i]) - y[i]);
      sd += w[i] * w[i];
    }

    /* Compute standard deviation */
    if (m > 1)
      sd = sqrt(sd / (double)(m - 1));
    else
      sd = sqrt(sd);

    if (sd == 0)
      /* The function fits the points perfectly */
      for (i = 0; i < m; i++)
	w[i] = 1;
    else
      /* Compute weight for each point */
      for (i = 0; i < m; i++){
	w[i] = w[i] / (_ITERATED_GLOBAL_NORMALISATION_SLOPE * sd);
	w[i] = pow(2, -w[i] * w[i]);
      }
  }
  else{
    /* Use function weight to assign weights to each point */
    /* Prepare computation of mean, standard deviation and median */
    for (i = 0; i < m; i++){
      w[i] = fabs(eval_function(f, x[i]) - y[i]);
      mean += w[i];
      sd += w[i] * w[i];
    }
    /* Compute mean, standard deviation and median */
    if (m > 0){
      mean = mean / m;
      if (m > 1)
	sd = sqrt(sd / (double)(m - 1));
      else
	sd = sqrt(sd);
      median = rankselect(w, (m - 1) / 2, m, sizeof(double),
			  (int (*)(void *, void *))cmp_doublep);
    }

    /* Compute weight for each point */
    for (i = 0; i < m; i++){
      w[i] = weight(w[i], mean, sd, median);
    }
  }

  return w;
}

/* ign(n, x, y, new_function, eval_function, cmp_functions,
 * free_function, weight): Compute normalising function by iteratively
 * applying new_function to a weighted version of the set of n points
 * given by x and y. The initial weighting assigns weight 1 to all
 * points, subsequent weights are computed by applying the function
 * weight to the distance of each point to the current normalising
 * function estimate, the mean distance, the standard deviation of the
 * distances, and the median of the distances for each point. If
 * weight is NULL, the weight is computed as 2 raised to the square of
 * the deviation of the point from the current normalising functions
 * measured in _IGN_SLOPE (see header file) standard deviations. Given
 * a return value f from new_function and a double x, eval_function
 * should return f(x). Given return values f and g from new_function,
 * cmp_functions should return True if f and g exhibits evidence of
 * convergence (e.g. are identical), False otherwise. Finally, given
 * return value f from new_function, free_function should free the
 * memory used by f.
 */
void *iterated_global_normalisation
(int n, double *x, double *y, _IGN_newfunction new_function,
 _IGN_evalfunction eval_function, _IGN_cmpfunction cmp_functions,
 _IGN_freefunction free_function, _IGN_weight weight)
{
  double *w = xmalloc(n * sizeof(double));
  void *f, *g;
  int i, k = 0, dif;

  /* Initialise normalising function to global normalisation function,
   * i.e. the normalising function obtained by setting all weights to 1.
   */
  for (i = 0; i < n; i++)
    w[i] = 1;

  f = new_function(n, x, y, w);
  /* Check existence of solution */
  if (f == NULL)
    /* An error occurred when trying to determine a normalising function */
    return NULL;

  /* Keep updating weights and normalising function until change - as
   * evaluated by cmp_functions - is negligible, or until a preset
   * maximum number of itereations is reached.
   */
  do {
    /* Compute new weights */
    wfree(w);
    w = weights(f, eval_function, n, x, y, weight);

    /* We need to remember previous function to keep track of improvement */
    g = f;

    /* Find new optimal normalising function */
    f = new_function(n, x, y, w);

    /* Check existence of solution */
    if (f == NULL){
      /* An error occurred when trying to determine a new normalising
       * function.
       */
      wfree(w);
      free_function(g);
      return NULL;
    }

    /* Compute difference between new and old normalising function */
    dif = !cmp_functions(f, g);

    /* All done with g - set it free */
    free_function(g);

    /* Increment iteration count */
    k++;
  } while (dif && (k < _ITERATED_GLOBAL_NORMALISATION_MAXITERATIONS));

  wfree(w);
  return f;
}
