/* polynomial.c: Implementation of functions to find the optimal
 * polynomial, in terms of least square difference, for a weighted set
 * of points.
 */
/* Rune Lyngs, January 2003 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "polynomial.h"
#include "wrappers.h"

/* xmalloc(n): Wrapper function for malloc checking return value */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "polynomial");
#else
  return wmalloc(n);
#endif
}

/* polynomial_new(n, c): returns a new polynomial with the n
 * coefficients given by c.
 */
Polynomial *polynomial_new(unsigned int n, double *a)
{
  Polynomial *p = (Polynomial *)xmalloc(sizeof(Polynomial));
  p->n = n;
  p->a = (double *)xmalloc(n * sizeof(double));
  for (;n > 0; n--)
    p->a[n - 1] = a[n - 1];

  return p;
}

/* invert(A, n): invert nn matrix A; return inverse, NULL if not invertible */
static double **invert(double **A, int n)
{
  int i, j, k;
  double **I = (double **)xmalloc(n * sizeof(double *));
  double **B = (double **)xmalloc(n * sizeof(double *));
  double tmp;

  /* Initialise I to identity matrix and copy A to B */
  for (i = 0; i < n; i++){
    I[i] = (double *)xmalloc(n * sizeof(double));
    B[i] = (double *)xmalloc(n * sizeof(double));
    for (j = 0; j < n; j++){
      I[i][j] = (i == j);
      B[i][j] = A[i][j];
    }
  }

  for (i = 0; i < n; i++){
    /* Handle column i */
    /* Find largest entry remaining in column */
    k = i;
    tmp = fabs(B[i][i]);
    for (j = i + 1; j < n; j++)
      if (fabs(B[j][i]) > tmp){
	tmp = fabs(B[j][i]);
	k = j;
      }
    /* Check for singular matrix */
    if (tmp == 0){
      fprintf(stderr, "Error: Singular matrix inversion attempted.\n");
      /* Free memory */
      for (i = 0; i < n; i++){
	wfree(I[i]);
	wfree(B[i]);
      }
      wfree(I);
      wfree(B);
      return NULL;
    }
    if (k != i){
      /* If it is not in the first remaining line, swap it there */
      for (j = i; j < n; j++){
	tmp = B[i][j];
	B[i][j] = B[k][j];
	B[k][j] = tmp;
      }
      for (j = 0; j < n; j++){
	tmp = I[i][j];
	I[i][j] = I[k][j];
	I[k][j] = tmp;
      }
    }

    for (j = 0; j < n; j++)
      if (j != i){
	/* Subtract row i from row j */
	tmp = B[j][i] / B[i][i];
	B[j][i] = 0;
	for (k = i + 1; k < n; k++)
	  B[j][k] = B[j][k] - tmp * B[i][k];
	for (k = 0; k < n; k++)
	  I[j][k] = I[j][k] - tmp * I[i][k];
      }
    /* Normalise row i */
    for (j = i + 1; j < n; j++)
      B[i][j] = B[i][j] / B[i][i];
    for (j = 0; j < n; j++)
      I[i][j] = I[i][j] / B[i][i];
    B[i][i] = 1.0;
  }

  /* Free memory */
  for (i = 0; i < n; i++)
    wfree(B[i]);
  wfree(B);

  /* Check stability */
  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++){
      tmp = A[i][0] * I[0][j];
      for (k = 1; k < n; k++)
	tmp += A[i][k] * I[k][j];
      if (fabs(tmp - (double)(i == j)) > _POLYNOMIAL_THRESHOLD){
	fprintf(stderr, "Warning: Unstable equation system [%f]\n",
		tmp - (double)(i == j));
	return I;
      }
    }

  return I;
}

/* equations(n, m, x, y, w, p, a, b): Create A and b for linear
 * equation system Ax = b yielding the optimal degree n - 1 polynomial
 * in the range from a to b with contribution (1 - p) from the sum of
 * square distances of the m points given by x and y weighted by w,
 * and contribution p from smoothness computed as integral of second
 * derivative squared from a to b. Return (n + 1)n matrix where first
 * n rows are a and last row is b.
 */
static double **equations(int n, int m, double *x, double *y, double *w,
			  double p, double a, double b)
{
  double **A;
  double *powx, powa, powb, *diffab;
  int i, j, column, row;

  /* Handle special cases */
  if (n < 1)
    return NULL;

  A = (double **)xmalloc((n + 1) * sizeof(double *));
  for (i = 0; i <= n; i++)
    A[i] = (double *)xmalloc(n * sizeof(double));

  if (n == 1){
    A[0][0] = 0;
    A[n][0] = 0;
    for (i = 0; i < m; i++){
      A[0][0] += w[i];
      A[n][0] += w[i] * y[i];
    }
    return A;
  }    

  if (p < 1){
    /* Compute contribution from SSD */
    /* For first cross diagonal of A and first entry of b we only need
     * the 0'th power of x, i.e. 1; simultaneously initialise powx.
     */
    powx = (double *)xmalloc(m * sizeof(double));
    A[0][0] = 0;
    A[n][0] = 0;
    for (i = 0; i < m; i++){
      powx[i] = x[i] * w[i];
      A[0][0] += w[i];
      A[n][0] += w[i] * y[i];
    }
    A[0][0] *= (1 - p);
    A[n][0] *= (1 - p);

    /* Compute rest of b vector and all cross diagonals of A but the
     * last one.
     */
    for (i = 1; i < 2 * n - 2; i++){
      if (i < n){
	/* Compute i'th entry of b vector */
	A[n][i] = 0;
	for (j = 0; j < m; j++)
	  A[n][i] += powx[j] * y[j];
	A[n][i] = (1 - p) * A[n][i];
	/* Compute first row and column of cross diagonal */
	column = 0;
	row = i;
      }
      else{
	/* We are past the main cross diagonal */
	column = i - n + 1;
	row = n - 1;
      }

      /* Compute A entries of this cross diagonal and next power of x values */
      A[row][column] = 0;
      for (j = 0; j < m; j++){
	A[row][column] += powx[j];
	powx[j] *= x[j];
      }
      A[row][column] *= (1 - p);
      /* Copy to rest of cross diagonal */
      for (j = 1; (j <= row) && (column + j < n); j++)
	A[row - j][column + j] = A[row][column];
    }

    /* Compute the last cross diagonal of A - we don't need to update
     * powx for next iteration, as this is the last cross diagonal.
     */
    A[n - 1][n - 1] = 0;
    for (j = 0; j < m; j++)
      A[n - 1][n - 1] += powx[j];
    A[n - 1][n - 1] *= (1 - p);

    /* Clean up */
    wfree(powx);
  }
  else
    for (i = 0; i <= n; i++)
      for (j = 0; j < n; j++)
	A[i][j] = 0;

  if ((p > 0) && (n > 2)){
    /* Compute contribution from smoothness */
    /* First compute difference between relevant powers of a and b */
    diffab = (double *)xmalloc((2 * n - 5) * sizeof(double));
    powa = a * p;
    powb = b * p;
    diffab[0] = powb - powa;
    for (i = 1; i < 2 * n - 5; i++){
      powa *= a;
      powb *= b;
      /* diffab[i] = b^(i + 1) - a^(i + 1) */
      diffab[i] = powb - powa;
    }

    /* Compute contributions to A from smoothness */
    for (i = 2; i < n; i++)
      for (j = 2; j < n; j++)
	A[i][j] += i * (i - 1) * j * (j - 1) / (double)(i + j - 3)
	  * diffab[i + j - 4];

    /* Clean up */
    wfree(diffab);
  }

  return A;
}

/* wlsdpolynomial(n, m, x, y, w, p, a, b): Find the polynomial of
 * degree n - 1 that minimises the sum of (1 - p) times the sum of
 * square differences to the weighted set of m points given by x (x
 * values), y (y values) and w (weights) and p times a smoothness term
 * computed as the integral from a to b of the square of the double
 * derivative of the polynomial, when weights have been normalised to
 * have average 1. Returns the array of the n coefficients (ordered
 * from a_0 to a_{n - 1}) of the polynomial, NULL if the polynomial is
 * not uniquely determined.
 */
Polynomial *polynomial_wlsdpolynomial(int n, int m, double *x, double *y,
				     double *w, double p, double a, double b)
{
  double **A, **I;
  double wsum = 0;
  int i, j;

  Polynomial *f = (Polynomial *)xmalloc(sizeof(Polynomial));
  f->n = n;
  f->a = (double *)xmalloc(n * sizeof(double));

  /* normalise weight (or, rather, modify p to emulate normalised weights */
  if ((p >= 0) && (p <= 1)){
    if ((p == 1) && (n > 2)){
      /* Only smoothness seems to be important; nevertheless, return
       * the straight line that fits the points best.
       */
      polynomial_free(f);
      return polynomial_wlsdpolynomial(2, m, x, y, w, 0, a, b);
    }
    else{
      if ((p > 0) && (p < 1)){
	/* Compute sum of weights */
	for (i = 0; i < m; i++)
	  wsum += w[i];
	/* Normalising weights is equivalent to rescaling p by average
	 * weight divided by 1 - p + p times average weight.
	 */
	p *= wsum / (m + (wsum - m) * p);
      }

      /* Set up equation system and find inverse */
      A = equations(n, m, x, y, w, p, a, b);
      I = invert(A, n);

      /* Check existence of unique solution */
      if (I == NULL){
	/* A is not invertible, i.e. the equation system is underdetermined */
	for (i = 0; i <= n; i++)
	  wfree(A[i]);
	wfree(A);
	polynomial_free(f);
	return NULL;
      }

      /* Compute solution from A^{-1} and b */
      for (i = 0; i < n; i++){
	f->a[i] = I[i][0] * A[n][0];
	for (j = 1; j < n; j++)
	  f->a[i] += I[i][j] * A[n][j];
      }

      /* Clean up */
      for (i = 0; i < n; i++){
	wfree(A[i]);
	wfree(I[i]);
      }
      wfree(A[n]);
      wfree(A);
      wfree(I);
    }
  }
  else{
    /* Smoothness parameter has to be in the range [0, 1] */
    fprintf(stderr, "Confounding weighting (%f) of smoothness requirement.\n",
	    p);
    polynomial_free(f);
    f = NULL;
  }

  return f;
}

/* lsdpolynomial(n, m, x, y, p, a, b): Find the polynomial of degree
 * n - 1 that minimises the sum of (1 - p) times the sum of square
 * differences to the unweighted set of m points given by x (x values)
 * and y (y values) and p times a smoothness term computed as the
 * integral from a to b of the square of the double derivative of the
 * polynomial. Returns the array of the n coefficients (ordered from
 * a_0 to a_{n - 1}) of the polynomial, NULL if the polynomial is not
 * uniquely determined.
 */
Polynomial *polynomial_lsdpolynomial(int n, int m, double *x, double *y,
				     double p, double a, double b)
{
  int i;
  double *w = (double *)xmalloc(m * sizeof(double));
  Polynomial *q;

  for (i = 0; i < m; i++)
    w[i] = 1;

  q = polynomial_wlsdpolynomial(n, m, x, y, w, p, a, b);
  wfree(w);

  return q;
}

/* fit(n, x, y): determines the unique polynomial of degree n - 1
 * agreeing with the n points given by x and y. If we can find n
 * polynomials p_1,...,p_n such that p_i(x_i) = 1 and p_i(x_j) = 0 for
 * j != i, then this polynomial is just \sum_i y_i * p_i. But p_i is
 * easily determined as \prod_{j != i} (x - x_j) / (x_i - x_j).
 * Acknowledgement: Thanks to Gerton Lunter for suggesting this.
 */
Polynomial *polynomial_fit(int n, double *x, double *y)
{
  int i, j, k;
  Polynomial *p = (Polynomial *)xmalloc(sizeof(Polynomial));
  double *q = (double *)xmalloc(n * sizeof(double));

  /* Initialise p */
  p->n = n;
  p->a = (double *)xmalloc(n * sizeof(double));
  for (i = 0; i < n; i++)
    p->a[i] = 0;

  for (i = 0; i < n; i++){
    /* Determine coefficients of p_i */
    q[n - 1] = 1;
    for (j = 0; j < n; j++){
      /* Skip if j = i */
      if (i == j) continue;
      /* Divide all coefficient with (x_i - x_j) */
      for (k = n - 1 - j + (i < j); k < n; k++)
	q[k] /= x[i] - x[j];
      /* Multiply polynomial in q with (x - x_j) */
      q[n - 2 - j + (i < j)] = -x[j] * q[n - 1 - j + (i < j)];
      for (k = n - 1 - j + (i < j); k < n - 1; k++)
	q[k] += -x[j] * q[k + 1];
    }
    /* Add p_i to p */
    for (j = 0; j < n; j++)
      p->a[j] += y[i] * q[j];
  }

  /* Clean up */
  wfree(q);

  return p;
}

/* product(p1, p2): compute polynomial that is the product of
 * polynomial p1 and p2.
 */
Polynomial *polynomial_product(Polynomial *p1, Polynomial *p2)
{
  Polynomial *q = (Polynomial *)xmalloc(sizeof(Polynomial));
  int i, j, n1, n2;

  /* Find highest order coefficient that is non zero for p1 */
  n1 = p1->n - 1;
  while ((n1 >= 0) && (p1->a[n1] == 0))
    n1--;
  /* Find highest order coefficient that is non zero for p2 */
  n2 = p1->n - 1;
  while ((n2 >= 0) && (p1->a[n2] == 0))
    n2--;

  if ((n1 < 0) || (n2 < 0))
    /* If either p1 or p2 is null polynomial, then so is p1 * p2 */
    q->n = 1;
  else
    /* Otherwise, degree of q is the sum of the degrees (which is one
     * less than the number of coefficients) of p1 and p2.
     */
    q->n = n1 + n2 + 1;

  /* Allocate memory for coefficients */
  q->a = (double *)xmalloc(q->n * sizeof(double));
  for (i = 0; i < q->n; i++)
    q->a[i] = 0;
  /* Perform muliplication */
  for (i = 0; i <= n1; i++)
    for (j = 0; j <= n2; j++)
      q->a[i + j] += p1->a[i] * p2->a[j];

  return q;
}

/* difference(p1, p2): compute polynomial p1 - p2 */
Polynomial *polynomial_difference(Polynomial *p1, Polynomial *p2)
{
  Polynomial *q = (Polynomial *)xmalloc(sizeof(Polynomial));
  int i, n1, n2;

  /* Find highest order coefficient that is non zero for p1, treating
   * the null polynomial as a polynomial of degree zero.
   */
  n1 = p1->n - 1;
  while ((n1 > 0) && (p1->a[n1] == 0))
    n1--;
  /* Find highest order coefficient that is non zero for p2, treating
   * the null polynomial as a polynomial of degree zero.
   */
  n2 = p1->n - 1;
  while ((n2 > 0) && (p1->a[n2] == 0))
    n2--;

  /* Degree (which is one less than the number of coefficients) of q
   * is equal to the maximum degree of p1 and p2.
   */
  q->n = (n1 > n2 ? n1 : n2) + 1;
  q->a = (double *)xmalloc(q->n * sizeof(double));
  /* Perform subtraction */
  for (i = 0; i < q->n; i++){
    q->a[i] = (i < p1->n ? p1->a[i] : 0);
    if (i < p2->n)
      q->a[i] -= p2->a[i];
  }

  return q;
}

/* integral(p, a, b): Compute integral of p from a to b */
double polynomial_integral(Polynomial *p, double a, double b)
{
  double powa = a, powb = b, i = p->a[0] * (b - a);
  int j;

  for (j = 1; j < p->n; j++){
    powa *= a;
    powb *= b;
    i += p->a[j] * (powb - powa) / (double)(j + 1);
  }
  return i;
}

/* evaluate(n, p, x): compute the value of degree n - 1 polynomial
 * with coefficients given by array p (ordered from a_0 to a_{n - 1})
 * at point x. Assumes n > 0.
 */
double polynomial_evaluate(Polynomial *p, double x)
{
  double powx, v;
  int i;

  /* Compute value of p at x */
  powx = x;
  v = p->a[0];
  for (i = 1; i < p->n; i++){
    v += p->a[i] * powx;
    powx *= x;
  }

  return v;
}

/* isd(p, q, a, b): compute integral of squared difference between p1
 * and p2 in interval from a to b.
 */
double polynomial_isd(Polynomial *p1, Polynomial *p2, double a, double b)
{
  Polynomial *p = polynomial_difference(p1, p2);
  Polynomial *q = polynomial_product(p, p);
  double isd = polynomial_integral(q, a, b);

  polynomial_free(p);
  polynomial_free(q);
  return isd;
}

/* print(p, format, fp): print polynomial p, using format to format
 * coefficients (default is to print each coefficient with three
 * digits after the decimal point), to file fp.
 */
void polynomial_print(Polynomial *p, char *format, FILE *fp)
{
  int i, j = p->n - 1;

  if (fp == NULL)
    fp = stdout;
  if (format == NULL)
    format = "%.3f";

  while ((j >= 0) && (p->a[j] == 0))
    j--;

  if (j < 0)
    fprintf(fp, "0\n");
  else{
    fprintf(fp, format, p->a[0]);
    if (j > 0){
      if (p->a[1] > 0){
	fprintf(fp, " + ");
	fprintf(fp, format, p->a[1]);
      }
      else if (p->a[1] < 0){
	fprintf(fp, " - ");
	fprintf(fp, format, -p->a[1]);
      }
      fprintf(fp, "x");
      for (i = 2; i <= j; i++){
	if (p->a[i] > 0){
	  fprintf(fp, " + ");
	  fprintf(fp, format, p->a[i]);
	}
	else{
	  fprintf(fp, " - ");
	  fprintf(fp, format, -p->a[i]);
	}
	fprintf(fp, "x^%d", i);
      }
    }
    fprintf(fp, "\n");
  }
}

void polynomial_free(Polynomial *p)
{
  wfree(p->a);
  wfree(p);
}
