/* splay.c: Implementation of splay trees */
/* Rune Lyngs, January 2003 */

#include <stdio.h>
#include <stdlib.h>
#include "wrappers.h"
#include "splay.h"

static splaynode *recycledsplaynodes = NULL;

/* xmalloc(n): Tries to allocate n bytes of memory; if unsuccessful an
 * error is reported.
 */
static void *xmalloc(size_t n)
{
#ifdef MPROFILING
  return wmalloc(n, "splay");
#else
  return wmalloc(n);
#endif
}

/* is_root(node): Returns true if node is root (i.e. if node does not
 * have a parent), false otherwise.
 */
static int is_root(splaynode *node)
{
  return node->parent == NULL;
}

/* is_sonofroot(node): Returns true if node is son of root; it is an
 * error to call this function with the root node, i.e. a node with no
 * parent.
 */
static int is_sonofroot(splaynode *node)
{
  return node->parent->parent == NULL;
}

/* is_leftson(node): Returns true if node is left son of its parent;
 * it is an error to call this function with the root node.
 */
static int is_leftson(splaynode *node)
{
  return node->parent->left == node;
}

/* is_rightson(node): Returns true if node is right son of its parent;
 * it is an error to call this function with the root node.
 */
static int is_rightson(splaynode *node)
{
  return node->parent->right == node;
}

/* inconsistent(): Issue inconsistency warning - by sending all
 * inconsitency detections to one place it is easy to set a break
 * point in gdb that catches the first inconsistency.
 */
static void inconsistent()
{
  fprintf(stderr, "Inconsistency detected\n");
}

/* check_consistency(t): Check consistency of splay tree t */
void check_consistency(splaytree *t)
{
  splaynode *n = t->root;

  while (n != NULL){
    if (n->left != NULL){
      /* As long as there is a left son, recurse on this */
      if (n->left->parent != n)
	inconsistent();
      n = n->left;
    }
    else if (n->right != NULL){
      /* As long as there is a right son, recurse on this */
      if (n->right->parent != n)
	inconsistent();
      n = n->right;
    }
    else
      /* We have reached a leaf - retrace to find the most recent
       * ancestor that is a left son and has a sibling.
       */
      while (!is_root(n)){
	if ((is_rightson(n)) || (n->parent->right == NULL))
	  n = n->parent;
	else{
	  if (n->parent != n->parent->right->parent)
	    inconsistent();
	  n = n->parent->right;
	  break;
	}
      }
    if (is_root(n)){
      if (n->parent != NULL)
	inconsistent();
      n = NULL;
    }
  }
}

/* rotate(node, update): Rotate node one level up and apply update
 * function (if different from NULL) to affected nodes (i.e. nodes
 * that have one or both subtrees changed); it is an error to call
 * this function with the root node.
 */
static void rotate(splaynode *node, void (*update)(void *, void *, void *))
{
  void *left, *right;

  if (is_leftson(node)){
    /* Move right subtree of node to left subtree of node's parent */
    node->parent->left = node->right;
    if (node->right != NULL)
      node->right->parent = node->parent;
    /* Make node's parent right child of node and set node's parent to
     * its previous grandparent.
     */
    node->right = node->parent;
    node->parent = node->parent->parent;
    if (!is_root(node->right)){
      if (is_leftson(node->right))
	node->parent->left = node;
      else
	node->parent->right = node;
    }
    node->right->parent = node;

    if (update != NULL){
      /* Update tree augmenting values for nodes affected by rotate */
      if (node->right->left != NULL)
	left = node->right->left->item;
      else
	left = NULL;
      if (node->right->right != NULL)
	right = node->right->right->item;
      else
	right = NULL;
      update(node->right->item, left, right);
      if (node->left != NULL)
	left = node->left->item;
      else
	left = NULL;
      update(node->item, left, node->right->item);
    }
  }
  else{
    /* Move left subtree of node to right subtree of node's parent */
    node->parent->right = node->left;
    if (node->left != NULL)
      node->left->parent = node->parent;
    /* Make node's parent left child of node and set node's parent to
     * its previous grandparent.
     */
    node->left = node->parent;
    node->parent = node->parent->parent;
    if (!is_root(node->left)){
      if (is_leftson(node->left))
	node->parent->left = node;
      else
	node->parent->right = node;
    }
    node->left->parent = node;

    if (update != NULL){
      /* Update tree augmenting values for nodes affected by rotate */
      if (node->left->left != NULL)
	left = node->left->left->item;
      else
	left = NULL;
      if (node->left->right != NULL)
	right = node->left->right->item;
      else
	right = NULL;
      update(node->left->item, left, right);
      if (node->right != NULL)
	right = node->right->item;
      else
	right = NULL;
      update(node->item, node->left->item, right);
    }
  }
}

/* macrostep(node): Perform a splay macrostep consisting of two
 * rotations, thus replacing the grandparent of node with node and
 * apply update to all affected nodes; it is an error to call this
 * function with a either the root node or one of the sons of the root
 * node.
 */
static void macrostep(splaynode *node, void (*update)(void *, void *, void *))
{
  void *t1, *t2, *t3, *t4;

  if (update != NULL){
    /* Prepare for tree augmenting values update */
    if (node->left != NULL)
      t1 = node->left->item;
    else
      t1 = NULL;
    if (node->right != NULL)
      t2 = node->right->item;
    else
      t2 = NULL;
  }

  if (is_leftson(node)){
    /* node is left son */
    if (update != NULL){
      /* Prepare for tree augmenting values update */
      if (node->parent->right != NULL)
	t3 = node->parent->right->item;
      else
	t3 = NULL;
    }
    /* In the two cases where node is a left son, the right subtree of
     * node is moved to the left subtree of node's parent and the
     * parent of node becomes its right son.
     */
    node->parent->left = node->right;
    if (node->right != NULL)
      node->right->parent = node->parent;
    node->right = node->parent;
    /* Set node's new parent relations */
    node->parent = node->parent->parent->parent;
    /* Note: node's old parent is now its right son */
    if (node->parent != NULL){
      if (is_leftson(node->right->parent))
	node->parent->left = node;
      else
	node->parent->right = node;
    }
    if (is_leftson(node->right)){
      /* node is left son of a left son */
      /* Move right subtree of node's parent to left subtree of node's
       * grandparent.
       */
      node->right->parent->left = node->right->right;
      if (node->right->right != NULL)
	node->right->right->parent = node->right->parent;
      /* Make node's grandparent the right son of node's parent */
      node->right->right = node->right->parent;
      node->right->right->parent = node->right;
      if (update != NULL){
	/* Prepare for tree augmenting values update */
	if (node->right->right->right != NULL)
	  t4 = node->right->right->right->item;
	else
	  t4 = NULL;
	/* Update tree augmenting values for nodes affected by macrostep */
	update(node->right->right->item, t3, t4);
	update(node->right->item, t2, node->right->right->item);
	update(node->item, t1, node->right->item);
      }
    }
    else{
      /* node is left son of a right son */
      /* Move left subtree of node to right subtree of node's grandparent */
      node->right->parent->right = node->left;
      if (node->left != NULL)
	node->left->parent = node->right->parent;
      /* Make node's grandparent the left son of node */
      node->left = node->right->parent;
      node->left->parent = node;
      if (update != NULL){
	/* Prepare for tree augmenting values update */
	if (node->left->left != NULL)
	  t4 = node->left->left->item;
	else
	  t4 = NULL;
	/* Update tree augmenting values for nodes affected by macrostep */
	update(node->left->item, t4, t1);
	update(node->right->item, t2, t3);
	update(node->item, node->left->item, node->right->item);
      }
    }
    /* Set the parent of node's old parent to node */
    node->right->parent = node;
  }
  else{
    /* node is right son */
    if (update != NULL){
      /* Prepare for tree augmenting values update */
      if (node->parent->left != NULL)
	t3 = node->parent->left->item;
      else
	t3 = NULL;
    }
    /* In the two cases where node is a right son, the left subtree of
     * node is moved to the right subtree of node's parent and the
     * parent of node becomes its left son.
     */
    node->parent->right = node->left;
    if (node->left != NULL)
      node->left->parent = node->parent;
    node->left = node->parent;
    /* Set node's new parent relations */
    node->parent = node->parent->parent->parent;
    /* Note: node's old parent is now its left son */
    if (node->parent != NULL){
      if (is_leftson(node->left->parent))
	node->parent->left = node;
      else
	node->parent->right = node;
    }
    if (is_rightson(node->left)){
      /* node is right son of a right son */
      /* Move left subtree of node's parent to right subtree of node's
       * grandparent.
       */
      node->left->parent->right = node->left->left;
      if (node->left->left != NULL)
	node->left->left->parent = node->left->parent;
      /* Make node's grandparent the left son of node's parent */
      node->left->left = node->left->parent;
      node->left->left->parent = node->left;
      if (update != NULL){
	/* Prepare for tree augmenting values update */
	if (node->left->left->left != NULL)
	  t4 = node->left->left->left->item;
	else
	  t4 = NULL;
	/* Update tree augmenting values for nodes affected by macrostep */
	update(node->left->left->item, t4, t3);
	update(node->left->item, node->left->left->item, t1);
	update(node->item, node->left->item, t2);
      }
    }
    else{
      /* node is right son of a left son */
      /* Move right subtree of node to left subtree of node's grandparent */
      node->left->parent->left = node->right;
      if (node->right != NULL)
	node->right->parent = node->left->parent;
      /* Make node's grandparent the right son of node */
      node->right = node->left->parent;
      node->right->parent = node;
      if (update != NULL){
	/* Prepare for tree augmenting values update */
	if (node->right->right != NULL)
	  t4 = node->right->right->item;
	else
	  t4 = NULL;
	/* Update tree augmenting values for nodes affected by macrostep */
	update(node->left->item, t3, t1);
	update(node->right->item, t2, t4);
	update(node->item, node->left->item, node->right->item);
      }
    }
    /* Set the parent of node's old parent to node */
    node->left->parent = node;
  }
}

/* createnode(item, left, right, parent): Create a new splay tree node
 * with item as item, left as left son, right as right son and parent
 * as parent.
 */
static splaynode *createnode(void *item, splaynode *left, splaynode *right,
			     splaynode *parent)
{
  splaynode *new;

#ifdef MPROFILING
  /* If the new node is used in the data structure keeping track of
   * used memory, it should not be taken from the set of recycled
   * nodes.
   */
  if (_inside_wrapper){
    new = (splaynode *)xmalloc(sizeof(splaynode));
    new->right = recycledsplaynodes;
    recycledsplaynodes = new;
  }
#endif

  /* Allocate memory for the new splaynode if a recycled spalynode is
   * not available.
   */
  if (recycledsplaynodes == NULL)
    new = (splaynode *)xmalloc(sizeof(splaynode));
  else{
    new = recycledsplaynodes;
    recycledsplaynodes = recycledsplaynodes->right;
  }

  /* Initialise splaynode */
  new->item = item;
  new->left = left;
  new->right = right;
  new->parent = parent;

  return new;
}

/* splay_preallocatenodes(n): creates n new nodes that are
 * inserted in the list of recycled nodes. We have to create each node
 * individually instead of just allocating the nodes as one big chunk
 * of memory as we would otherwise run into problems when trying to
 * free individual nodes. So the only reason there might be for
 * preallocating nodes is to (hopefully) have the nodes located
 * consecutively in memory, thus perhaps yielding a better cache
 * performance.
 */
void splay_preallocatenodes(int n)
{
  int i;
  splaynode *tmp;

  for (i = 0; i < n ; i++){
    tmp = (splaynode *)xmalloc(sizeof(splaynode));
    tmp->right = recycledsplaynodes;
    recycledsplaynodes = tmp;
  }
}

/* splay_freeunusednodes(): Frees memory of all nodes currently in the
 * list of recycled nodes.
 */
void splay_freeunusednodes()
{
  splaynode *tmp = recycledsplaynodes;

  while (recycledsplaynodes != NULL){
    tmp = recycledsplaynodes->right;
    wfree(recycledsplaynodes);
    recycledsplaynodes = tmp;
  }
}

/* void destroysplaynode(node): frees the memory used by node.
 */
static void destroysplaynode(splaynode *node)
{
  wfree(node);
}

/* deletesplaynode(node): inserts node at the front of the list of
 * recycled nodes.
 */
static void deletesplaynode(splaynode *node)
{
#ifdef MPROFILING
  /* If the node being deleted was used in the data structure keeping
   * track of allocated memory it should not be recycled.
   */
  if (_inside_wrapper){
    destroysplaynode(node);
    return;
  }
#endif

  node->right = recycledsplaynodes;
  recycledsplaynodes = node;
}

/* void deletetree(node, delete_item): recycles all the nodes in the
 * subtree rooted at node. Note: the parent of node, or the root
 * pointer of the tree containing node if node is the root, is not
 * updated to reflect the deletion of node (and all its
 * descendants). If not NULL the function delete_item is applied to
 * all items stored in the subtree being deleted.
 */
static void deletetree(splaynode *node, void (*delete_item)(void *))
{
  if (node != NULL){
    /* Apply delete_item to item of current node */
    if (delete_item != NULL)
      delete_item(node->item);
    /* Recursively delete left subtree */
    deletetree(node->left, delete_item);
    /* Recursively delete right subtree */
    deletetree(node->right, delete_item);
    /* Finally delete the current node */
    deletesplaynode(node);
  }
}

/* void destroytree(node, destroy_item): frees the memory used by all
 * nodes in the subtree rooted at node. Note: the parent of node, or
 * the root pointer of the tree containing node if node is the root,
 * is not updated to reflect the destruction of node (and all its
 * descendants). If not NULL the function destroy_item is applied to
 * all items stored in the subtree being deleted.
 */
static void destroytree(splaynode *node, void (*destroy_item)(void *))
{
  if (node != NULL){
    /* Apply destroy_item to item of current node */
    if (destroy_item != NULL)
      destroy_item(node->item);
    /* Recursively destroy left subtree */
    destroytree(node->left, destroy_item);
    /* Recursively destroy right subtree */
    destroytree(node->right, destroy_item);
    /* Finally destroy the current node */
    destroysplaynode(node);
  }
}

/* splay_deletetree(tree, delete_item): recycles all the nodes in the
 * tree and applies delete_item (if not NULL) to all items stored in
 * the tree.
 */
void splay_deletetree(splaytree *tree, void (*delete_item)(void *))
{
  deletetree(tree->root, delete_item);
  wfree(tree);
}

/* splay_destroytree: frees the memory used by all nodes in the tree
 * and applies destroy_item (if not NULL) to all items stored in the
 * tree.
 */
void splay_destroytree(splaytree *tree, void (*destroy_item)(void *))
{
  destroytree(tree->root, destroy_item);
  wfree(tree);
}

/* find(less_than, item, root): finds item in the tree rooted at root
 * using less_than for comparison. Assumption: the items in the tree
 * rooted at root are ordered according to less_than. Returns true if
 * item is in the tree rooted at root, false otherwise, and updates
 * the content of root to the address of the splaynode where the
 * search terminated.
 */
static int find(int (*less_than)(void *, void *), void *item, splaynode **root)
{
  /* Empty tree is a special case */
  if (*root == NULL)
    return 0;

  /* Find largest node not larger than item or smallest node not
   * smaller than item.
   */
  for (;;){
    if (less_than(item, (*root)->item)){
      /* item should be in left subtree of current */
      if ((*root)->left == NULL)
	/* item is not in the tree */
	return 0;

      *root = (*root)->left;
    }
    else if (less_than((*root)->item, item)){
      /* item should be in right subtree of current */
      if ((*root)->right == NULL)
	/* item is not in the tree */
	return 0;
      *root = (*root)->right;
    }
    else
      /* item in current equals item - current is the one */
      return 1;
  }
}

/* rotatetotop(node, update): move node up as root by repeated
 * rotations and apply update (if not NULL) to all affected
 * nodes. Note: the tree containing node is not updated to point to
 * node as its root.
 */
static void rotatetotop(splaynode *node,
			void (*update)(void *, void *, void *))
{
  while (!is_root(node)){
    if (is_sonofroot(node)){
      /* node is son of root - one more rotation and we are done */
      rotate(node, update);
      break;
    }
    else
      /* node is at least two levels down from the root - perform a
       * macro step
       */
      macrostep(node, update);
  }
}

/* splay(tree, item): find the largest item in tree that is not
 * larger than item or the smallest item in tree that is not smaller
 * than item and rotate it up as root in tree. Return true if the item
 * is in the tree, false otherwise.
 */
static int splay(splaytree *tree, void *item)
{
  int member = 0;

  /* Empty tree is a special case */
  if (tree->root == NULL)
    return member;

  /* Search for item in splaytree */
  member = find(tree->less_than, item, &tree->root);

  /* Rotate the node where the search was terminated to the top
   * (tree->root already points to this node as we used this variable
   * in the preceeding find).
   */
  rotatetotop(tree->root, tree->update);

  return member;
}

/* splaymin(node, update): rearrange the tree rooted at node as if we
 * had splayed it with the infinitely small element under the
 * less_than ordering, and apply update (if not NULL) to all affected
 * nodes. Return the root of the rearranged tree with parent set to
 * NULL.
 */
static splaynode *splaymin(splaynode *node,
			   void (*update)(void *, void *, void *))
{
  /* Cut off subtree rooted at node */
  node->parent = NULL;

  /* Find node with smallest item in tree rooted at node */
  while (node->left != NULL)
    node = node->left;

  /* Move it up as root */
  rotatetotop(node, update);

  return node;
}

/* splaymax(node): rearrange the tree rooted at node as if we had
 * splayed it with the infinitely large element under the less_than
 * ordering, and apply update (if not NULL) to all affected
 * nodes. Return the root of the rearranged tree with parent set to
 * NULL.
 */
static splaynode *splaymax(splaynode *node,
			   void (*update)(void *, void *, void*))
{
  /* Cut off subtree rooted at node */
  node->parent = NULL;

  /* Find node with largest item in tree rooted at node */
  while (node->right != NULL)
    node = node->right;

  /* Move it up as root */
  rotatetotop(node, update);

  return node;
}

/* splay_create(less_than): Create an empty splay tree with less_than
 * as comparison function and update as augmenting values updating
 * function.
 */
splaytree *splay_create(int (*less_than)(void *, void *),
			void (*update)(void *, void *, void *))
{
  splaytree *tree = (splaytree *)xmalloc(sizeof(splaytree));

  tree->root = NULL;
  tree->less_than = less_than;
  tree->update = update;

  return tree;
}

/* splay_insert(tree, item): Inserts item in tree. */
void splay_insert(splaytree *tree, void *item)
{
  void *t1, *t2;

  if (tree->root == NULL){
    /* Insertion in empty tree - create new node and insert as root */
    tree->root = createnode(item, NULL, NULL, NULL);
    if (tree->update != NULL)
      tree->update(item, NULL, NULL);
  }
  else{
    /* Insertion in non-empty tree - splay and insert item as root of
     * the two trees obtained by cutting of either the left or right
     * subtree of the root, depending on whether the root item is
     * smaller or larger than item.
     */
    splay(tree, item);
    if (tree->less_than(item, tree->root->item)){
      /* New item is less than root of splayed tree */
      tree->root = createnode(item, tree->root->left, tree->root, NULL);
      /* Cut off left subtree of the old root */
      tree->root->right->left = NULL;
      /* Update parent pointers of the sons of the new root */
      tree->root->right->parent = tree->root;
      if (tree->root->left != NULL){
	tree->root->left->parent = tree->root;
	t1 = tree->root->left->item;
      }
      else
	t1 = NULL;
      if (tree->update != NULL){
	/* Prepare for tree augmenting values update */
	if (tree->root->right->right != NULL)
	  t2 = tree->root->right->right->item;
	else
	  t2 = NULL;
	/* Update augmenting values for affected nodes */
	tree->update(tree->root->right->item, NULL, t2);
	tree->update(tree->root->item, t1, tree->root->right->item);
      }
    }
    else if (tree->less_than(tree->root->item, item)){
      /* New item is larger than root of splayed tree */
      tree->root = createnode(item, tree->root, tree->root->right, NULL);
      /* Cut off right subtree of the old root */
      tree->root->left->right = NULL;
      /* Update parent pointers of the sons of the new root */
      tree->root->left->parent = tree->root;
      if (tree->root->right != NULL){
	tree->root->right->parent = tree->root;
	t2 = tree->root->right->item;
      }
      else
	t2 = NULL;
      if (tree->update != NULL){
	/* Prepare for tree augmenting values update */
	if (tree->root->left->left != NULL)
	  t1 = tree->root->left->left->item;
	else
	  t1 = NULL;
	/* Update augmenting values for affected nodes */
	tree->update(tree->root->left->item, t1, NULL);
	tree->update(tree->root->item, tree->root->left->item, t2);
      }
    }
  }
}

/* deletenode(tree, node): delete node from tree and return a pointer
 * to the deleted item (NULL if the item was not present in the tree.
 */
static void *deletenode(splaytree *tree, splaynode *node)
{
  splaynode *new_root;
  void *t1;
  void *ritem;

  /* Standard way would be just to rotate node up as root, remove
   * it, and join the two resulting subtrees. But more than half the
   * nodes in a tree will have less than two subtrees. For these
   * nodes we can just remove node, move any subtree of node up as
   * subtree of node's parent, and rotate the parent of node up as
   * root.
   */
  if (node->left == NULL){
    if (is_root(node)){
      /* node is root and has no left subtree - just delete it and
       * promote its right subtree to full splay tree.
       */
      tree->root = node->right;
      if (tree->root != NULL)
	tree->root->parent = NULL;
    }
    else{
      /* node is somewhere down in the tree and has no left subtree;
       * delete it and move its right subtree up as subtree of
       * parent and rotate parent up as root.
       */
      if (is_leftson(node))
	node->parent->left = node->right;
      else
	node->parent->right = node->right;

      if (node->right != NULL)
	node->right->parent = node->parent;
      rotatetotop(node->parent, tree->update);
      tree->root = node->parent;
    }
  }
  else if (node->right == NULL){
    /* We now know that node has a non-empty left subtree; this
     * allows us to skip a few checks compared to the above code.
     */
    if (is_root(node)){
      /* node is root and has no right subtree - just delete it and
       * promote its left subtree to full splay tree.
       */
      tree->root = node->left;
      tree->root->parent = NULL;
    }
    else{
      /* node is somewhere down in the tree and has no right
       * subtree; delete it and move its left subtree up as subtree
       * of parent and rotate parent up as root.
       */
      if (is_leftson(node))
	node->parent->left = node->left;
      else
	node->parent->right = node->left;

      node->left->parent = node->parent;
      rotatetotop(node->parent, tree->update);
      tree->root = node->parent;
    }
  }
  else{
    /* node has both left and right subtrees - rotate it up as root,
     * delete it, and join its two subtrees.
     */
    rotatetotop(node, tree->update);
    new_root = splaymax(node->left, tree->update);
    new_root->right = node->right;
    new_root->right->parent = new_root;
    tree->root = new_root;
    if (tree->update != NULL){
      /* Prepare for update of tree augmenting values */
      if (new_root->left != NULL)
	t1 = new_root->left->item;
      else
	t1 = NULL;
      /* Update tree augmenting values for affected nodes */
      tree->update(new_root->item, t1, new_root->right->item);
    }
  }

  /* As we in all cases delete the node containing item, its actual
   * deletion has been postponed to now.
   */
  ritem = node->item;
  deletesplaynode(node);

  return ritem;
}

/* splay_delete(tree, item): delete item from tree and return a
 * pointer to the deleted item (NULL if the item was not present in
 * the tree).
 */
void *splay_delete(splaytree *tree, void *item)
{
  splaynode *node = tree->root;

  if (find(tree->less_than, item, &node)){
    /* Delete node from tree */
    return deletenode(tree, node);
  }

  /* No item is deleted, but we still need to rotate the node where
   * the search terminated to the top to get good amortised behaviour.
   */
  rotatetotop(node, tree->update);
  tree->root = node;

  return NULL;
}

/* splay_deleterange(tree, item1, less_than1, item2, less_than2,
 * delete_item): deletes all items in tree that are not less than
 * item1 using less_than1 for comparison, and are less than or equal
 * to item2 using less_than2 for comparison. All deleted items are
 * processed by delete_item (if delete_item is not NULL). Consistency
 * of less_than1, less_than2, and the ordering of the tree is assumed.
 */
void splay_deleterange(splaytree *tree,
		       void *item1, int (*less_than1)(void *, void *),
		       void *item2, int (*less_than2)(void *, void *),
		       void (*delete_item)(void *))
{
  splaynode *branchpoint, *current, *tmp, *nextnode;
  int leftson;

  /* Nothing to do for an empty tree */
  if (tree->root == NULL)
    return;

  /* Default comparison function is the one associated with the tree */
  if (less_than1 == NULL)
    less_than1 = tree->less_than;
  if (less_than2 == NULL)
    less_than2 = tree->less_than;

  /* if item2 is less than item1, the range is empty */
  if (less_than1(item2, item1))
    return;

  nextnode = tree->root;
  branchpoint = NULL;

  /* Find the branching point, i.e. the node that has item1 in the
   * left subtree and item2 in the right subtree.
   */
  for (;;){
    /* If we reach a leaf that is not a branching point, the range is empty */
    if (nextnode == NULL){
      /* But we still need to move the node where we ended the search
       * to the top to get good amortised behaviour.
       */
      rotatetotop(branchpoint, tree->update);
      tree->root = branchpoint;
      return;
    }

    /* The branchpoint is somewhere in the subtree rooted at nextnode */
    branchpoint = nextnode;

    /* We just need to compare the two items to the item of
     * branchpoint. Assuming that the ordering of the nodes in the
     * tree is consistent with both less_than1 and less_than2, if
     * item1 is in the right subtree then so is item2, and if item2 is
     * in the left subtree then so is item1; otherwise, branchpoint is
     * indeed the branching point.
     */
    if (less_than1(branchpoint->item, item1))
      /* Both items are in the right subtree of branchpoint */
      nextnode = branchpoint->right;
    else if (less_than2(item2, branchpoint->item))
      /* Both items are in the left subtree of branchpoint */
      nextnode = branchpoint->left;
    else
      /* We have reached the branching point */
      break;
  }

  if (less_than1(item1, branchpoint->item)){
    /* The item of the branchpoint is larger than item1. Remove all
     * elements not less than item1 in left subtree of the branching
     * point.
     */
    current = branchpoint;
    nextnode = branchpoint->left;
    leftson = 1;

    while (nextnode != NULL){
      if (less_than1(nextnode->item, item1)){
	/* All elements at least as large as item1 are in the right
	 * subtree of nextnode.
	 */
	current = nextnode;
	nextnode = nextnode->right;
	leftson = 0;
      }
      else{
	/* All elements smaller than item1 are in the left subtree of
	 * nextnode, i.e. all elements in the right subtree as well as
	 * nextnode should be deleted.
	 */
	/* Delete right subtree */
	deletetree(nextnode->right, delete_item);

	/* Delete nextnode */
	if (leftson)
	  current->left = nextnode->left;
	else
	  current->right = nextnode->left;
	if (nextnode->left != NULL)
	  nextnode->left->parent = current;

	if (less_than1(item1, nextnode->item)){
	  /* The item in nextnode was larger than item1, so prepare
	   * recursion on left subtree.
	   */
	  if (delete_item != NULL)
	    delete_item(nextnode->item);
	  tmp = nextnode->left;
	  deletesplaynode(nextnode);
	  nextnode = tmp;
	}
	else{
	  /* The item in nextnode was identical to item1, so we can stop
	   * the recursion.
	   */
	  if (delete_item != NULL)
	    delete_item(nextnode->item);
	  deletesplaynode(nextnode);
	  break;
	}
      }
    }

    /* Move the node where the search ended up as root of the tree;
     * the tree augmenting values for the nodes affected should be
     * updated during this operation.
     */
    /* It should be safe to perform this operation before deleting items
     * in the right subtree of branchingpoint, but if behaviour is
     * unexpected it might be a good idea to start debugging with moving
     * the splaying of current down below the following code that
     * deletes items in the right subtree of branchingpoint.
     */
    rotatetotop(current, tree->update);
    tree->root = current;
  }

  if (less_than2(branchpoint->item, item2)){
    /* The item of the branchpoint is smaller than item1. Remove all
     * elements not larger than item2 in right subtree of the
     * branching point.
     */
    current = branchpoint;
    nextnode = branchpoint->right;
    leftson = 0;

    while (nextnode != NULL){
      if (less_than2(item2, nextnode->item)){
	/* All elements no larger than item2 are in the left subtree of
	 * nextnode.
	 */
	current = nextnode;
	nextnode = nextnode->left;
	leftson = 1;
      }
      else{
	/* All elements larger than item2 are in the right subtree of
	 * nextnode, i.e. all elements in the left subtree as well as
	 * nextnode should be deleted.
	 */
	/* Delete left subtree */
	deletetree(nextnode->left, delete_item);

	/* Delete nextnode */
	if (leftson)
	  current->left = nextnode->right;
	else
	  current->right = nextnode->right;
	if (nextnode->right != NULL)
	  nextnode->right->parent = current;
	if (less_than2(nextnode->item, item2)){
	  /* The item in nextnode was smaller than item2, so prepare
	   * recursion on right subtree.
	   */
	  if (delete_item != NULL)
	    delete_item(nextnode->item);
	  tmp = nextnode->right;
	  deletesplaynode(nextnode);
	  nextnode = tmp;
	}
	else{
	  /* The item in nextnode was identical to item2, so we can stop
	   * the recursion.
	   */
	  if (delete_item != NULL)
	    delete_item(nextnode->item);
	  deletesplaynode(nextnode);
	  break;
	}
      }
    }
    /* Move the node where the search ended up as root of the tree;
     * the tree augmenting values for the nodes affected should be
     * updated during this operation.
     */
    rotatetotop(current, tree->update);
    tree->root = current;
  }

  /* We still need to delete the branching point */
  if (delete_item != NULL)
    delete_item(branchpoint->item);
  deletenode(tree, branchpoint);
}

/* splay_member(tree, item): returns true if item is in tree, false
 * otherwise.
 */
int splay_member(splaytree *tree, void *item)
{
  return splay(tree, item);
}

/* splay_access(tree, item): returns a pointer to item; if item is not
 * in tree, return NULL.
 */
void *splay_access(splaytree *tree, void *item)
{
  if (splay(tree, item))
    return tree->root->item;
  else
    return NULL;
}

/* splay_accesssmaller(tree, item): returns a pointer to largest item
 * in the tree no larger than item; if all items in the tree are
 * larger than item, returns NULL.
 */
void *splay_accesssmaller(splaytree *tree, void *item)
{
  if (tree->root == NULL)
    return NULL;

  if (!splay(tree, item) && tree->less_than(item, tree->root->item)){
    /* Root item is larger than item */
    if (tree->root->left == NULL)
      return NULL;
    else
      rotate(tree->root->left, tree->update);
  }

  return tree->root->item;
}

/* splay_accesslarger(tree, item): returns a pointer to smallest item
 * in the tree no smaller than item; if all items in the tree are
 * smaller than item, returns NULL.
 */
void *splay_accesslarger(splaytree *tree, void *item)
{
  if (tree->root == NULL)
    return NULL;

  if (!splay(tree, item) && tree->less_than(tree->root->item, item)){
    /* Root item is smaller than item */
    if (tree->root->right == NULL)
      return NULL;
    else
      rotate(tree->root->right, tree->update);
  }

  return tree->root->item;
}

/* splay_evaluate(tree, item, combine): Evaluate the augmenting values
 * of tree relative to item, using combine to combine augmenting
 * values of elements smaller than item, equal to item, and larger
 * than item.
 */
void *splay_evaluate(splaytree *tree, void *item,
		     void *(*combine)(void *, void *, void *))
{
  void *smaller = NULL, *equal = NULL, *larger = NULL, *value;

  if (tree->root == NULL)
    return combine(NULL, NULL, NULL);

  if (splay(tree, item))
    equal = tree->root->item;
  if (tree->root->left != NULL)
    smaller = tree->root->left->item;
  if (tree->root->right != NULL)
    larger = tree->root->right->item;
  if (equal != NULL)
    value = combine(smaller, equal, larger);
  else{
    if (tree->less_than(item, tree->root->item)){
      /* Smaller items are in the left subtree, larger in the root and
       * the right subtree; update augmenting value in root as if
       * there was no left subtree.
       */
      tree->update(tree->root->item, NULL, larger);
      value = combine(smaller, NULL, tree->root->item);
    }
    else{
      /* Smaller items are in the left subtree and the root, larger
       * the right subtree; update augmenting value in root as if
       * there was no right subtree.
       */
      tree->update(tree->root->item, smaller, NULL);
      value = combine(tree->root->item, NULL, larger);
    }
    /* Restore augmenting value in root */
    tree->update(tree->root->item, smaller, larger);
  }

  return value;
}


/* splay_join(tree1, tree2): joins the two trees. Assumptions: The two
 * trees have compatible orderings (and augmenting values) and all
 * items in tree1 are smaller than any item in tree2. The ordering of
 * the joined tree will be that of tree1. Tree2 is deleted.
 */
splaytree *splay_join(splaytree *tree1, splaytree *tree2)
{
  void *t1;

  /* If either of the trees is empty the join can be carried out by
   * just returning the other tree.
   */
  if (tree1->root != NULL){
    if (tree2->root != NULL){
      /* Move the maximum element of tree1 up as root; this makes the
       * root of tree1 have an empty right subtree, hence we can just
       * insert tree2 as right subtree of the root of tree1.
       */
      tree1->root = splaymax(tree1->root, tree1->update);
      tree1->root->right = tree2->root;
      tree2->root->parent = tree1->root;
      if (tree1->update != NULL){
	/* Prepare for update of tree augmenting values */
	if (tree1->root->left != NULL)
	  t1 = tree1->root->left->item;
	else
	  t1 = NULL;
	/* Update tree augmenting values for affected nodes */
	tree1->update(tree1->root->item, t1, tree1->root->right->item);
      }
    }
  }
  else
    tree1->root = tree2->root;

  wfree(tree2);

  return tree1;
}

/* splay_split(tree, item): removes all items larger than item from
 * tree and returns a new splay tree with these items and the same
 * ordering as tree.
 */
splaytree *splay_split(splaytree *tree, void *item)
{
  void *t1;
  splaytree *new_tree = (splaytree *)xmalloc(sizeof(splaytree));
  new_tree->less_than = tree->less_than;
  new_tree->update = tree->update;

  if (tree->root == NULL){
    /* The tree is empty - simply return another empty tree */
    new_tree->root = NULL;
  }
  else{
    /* Rearrange tree so that the node containing the largest item not
     * larger than item or smallest item not smaller than item becomes
     * root.
     */
    splay(tree, item);
    if (tree->less_than(item, tree->root->item)){
      /* New root is larger than item - replace tree with its left
       * subtree and move the remaining elements to a new, independent
       * splay tree.
       */
      new_tree->root = tree->root;
      tree->root = tree->root->left;
      if (tree->root != NULL){
	new_tree->root->left = NULL;
	tree->root->parent = NULL;
	if (tree->update != NULL){
	  /* Prepare for update of augmenting values */
	  if (new_tree->root->right != NULL)
	    t1 = new_tree->root->right->item;
	  else
	    t1 = NULL;
	  /* Update augmenting values for affected nodes */
	  tree->update(new_tree->root->item, NULL, t1);
	}
      }
    }
    else{
      /* New root is no larger than item - cut off right subtree and
       * make it a new independent splay tree.
       */
      new_tree->root = tree->root->right;
      if (new_tree->root != NULL){
	new_tree->root->parent = NULL;
	tree->root->right = NULL;
	if (tree->update != NULL){
	  /* Prepare for update of augmenting values */
	  if (tree->root->left != NULL)
	    t1 = tree->root->left->item;
	  else
	    t1 = NULL;
	  /* Update augmenting values for affected nodes */
	  tree->update(tree->root->item, t1, NULL);
	}
      }
    }
  }

  return new_tree;
}

/* splay_min(tree): returns the smallest item in tree */
void *splay_min(splaytree *tree)
{
  if (tree->root == NULL)
    /* Empty tree - just return NULL */
    return NULL;

  tree->root = splaymin(tree->root, tree->update);

  return tree->root->item;
}

/* splay_max(tree): returns the largest item in tree */
void *splay_max(splaytree *tree)
{
  if (tree->root == NULL)
    /* Empty tree - just return NULL */
    return NULL;

  tree->root = splaymax(tree->root, tree->update);

  return tree->root->item;
}

/* splay_deletemin(tree): deletes the smallest item in tree and returns it */
void *splay_deletemin(splaytree *tree)
{
  void *item = splay_min(tree);

  if (tree->root != NULL){
    if (tree->root->right != NULL){
      /* Tree does not become empty by delete operation */
      tree->root = tree->root->right;
      deletesplaynode(tree->root->parent);
      tree->root->parent = NULL;
    }
    else{
      /* Only one item left in tree */
      deletesplaynode(tree->root);
      tree->root = NULL;
    }
  }

  return item;
}

/* splay_deletemax(tree): deletes the largest item in tree and returns it */
void *splay_deletemax(splaytree *tree)
{
  void *item = splay_max(tree);

  if (tree->root != NULL){
    if (tree->root->left != NULL){
      /* Tree does not become empty by delete operation */
      tree->root = tree->root->left;
      deletesplaynode(tree->root->parent);
      tree->root->parent = NULL;
    }
    else{
      /* Only one item left in tree */
      deletesplaynode(tree->root);
      tree->root = NULL;
    }
  }

  return item;
}

/* splay_empty(tree): returns true of the tree is empty, false otherwise */
int splay_empty(splaytree *tree)
{
  return (tree->root == NULL);
}

/* printsubtree(node, print_item): prints the subtree rooted at
 * node with parentheses specifying the tree structure, using the
 * function print_item to print items.
 */
static void printsubtree(splaynode *node, void (*print_item)(void *))
{
  if (node->left != NULL){
    /* print left subtree */
    printf("(");
    printsubtree(node->left, print_item);
    printf(")");
  }
  print_item(node->item);
  if (node->right != NULL){
    /* print right subtree */
    printf("(");
    printsubtree(node->right, print_item);
    printf(")");
  }
}

/* splay_printtree(tree, print_item): prints the tree with parentheses
 * specifying the tree structure, using the function print_item to
 * print items.
 */
void splay_printtree(splaytree *tree, void (*print_item)(void *))
{
  if (tree->root != NULL)
    printsubtree(tree->root, print_item);
  printf("\n");
}

/* listsubtreeelements(node, print_item, separator): prints the
 * subtree rooted at node using print_item to print items and printing
 * separator between all printed elements.
 */
void listsubtreeelements(splaynode *node, void (*print_item)(void *),
			 char *separator)
{
  if (node->left != NULL){
    /* list elements of left subtree */
    listsubtreeelements(node->left, print_item, separator);
    if (separator != NULL)
      printf(separator);
  }
  print_item(node->item);
  if (node->right != NULL){
    /* list elements of right subtree */
    if (separator != NULL)
      printf(separator);
    listsubtreeelements(node->right, print_item, separator);
  }
}
    
/* splay_listtreeelements(tree, print_item, separator): prints the
 * elements of the tree (without the added structural information
 * printed by printtree) using print_item to print items, with
 * separator between all printed elements.
 */
void splay_listtreeelements(splaytree *tree, void (*print_item)(void *),
			char *separator)
{
  if (tree->root != NULL)
    listsubtreeelements(tree->root, print_item, separator);
  printf("\n");
}
